import logging
from reliquery.metadata import Metadata, MetadataDB, RelicData, RelicTag
from typing import List, Dict
from sys import getsizeof
from io import BytesIO

import numpy as np
import json
import pandas as pd

import nbconvert
import nbformat

from .storage import (
    get_all_available_storages,
    StorageItemDoesNotExist,
    get_storage_by_name,
    Storage,
)


StoragePath = List[str]

dt_format = "%m/%d/%Y %H:%M:%S"


class InvalidRelicId(Exception):
    pass


class EmptyTagDict(Exception):
    pass


class Relic:
    name: str
    relic_type: str
    storage: Storage

    def __init__(
        self,
        name: str,
        relic_type: str,
        storage: Storage = None,
        storage_name: str = "default",
        check_exists: bool = True,
    ):
        self.name = name
        self.relic_type = relic_type
        self.storage_name = storage_name

        if storage is None:
            self.storage = get_storage_by_name(self.storage_name)

        else:
            self.storage = storage

        if check_exists:
            self._ensure_exists()

        self.metadata_db = MetadataDB()
        self.data = self.metadata_db.sync_relic_data(
            RelicData(self.name, self.relic_type, self.storage_name)
        )

    @classmethod
    def assert_valid_id(cls, id: str):
        if id == "":
            raise InvalidRelicId()

    def _ensure_exists(self):
        try:
            self.storage.get_text([self.relic_type, self.name, "exists"])
        except StorageItemDoesNotExist:
            logging.info("Creating a Relic")
            self.storage.put_text([self.relic_type, self.name, "exists"], "exists")

    # TODO: needs test coverage
    @classmethod
    def relic_exists(
        cls,
        name: str,
        relic_type: str,
        storage: Storage = None,
        storage_name: str = "default",
    ) -> bool:
        if storage is None:
            storage = get_storage_by_name(storage_name)
        try:
            storage.get_text([relic_type, name, "exists"])
        except StorageItemDoesNotExist:
            return False

        return True

    def _relic_data(self):
        return self.metadata_db.get_relic_data_by_name(
            self.name, self.relic_type, self.storage_name
        )

    def add_array(self, name: str, array: np.ndarray):

        self.assert_valid_id(name)

        size = getsizeof(array)
        shape = str(np.array(array).shape)
        metadata = Metadata(
            name=name,
            data_type="arrays",
            relic=self.data,
            size=size,
            shape=shape,
        )

        buffer = BytesIO()
        np.save(buffer, array, allow_pickle=False)
        buffer.seek(0)

        self.storage.put_binary_obj(
            [self.relic_type, self.name, "arrays", name], buffer
        )
        self._add_metadata(metadata)

    def get_array(self, name: str) -> np.ndarray:
        self.assert_valid_id(name)

        with self.storage.get_binary_obj(
            [self.relic_type, self.name, "arrays", name]
        ) as f:
            return np.load(f, allow_pickle=False)

    def list_arrays(self) -> List[np.ndarray]:
        return self.storage.list_keys([self.relic_type, self.name, "arrays"])

    def add_html_from_path(self, name: str, html_path: str) -> None:

        self.assert_valid_id(name)

        metadata = Metadata(name=name, data_type="html", relic=self.data)

        self.storage.put_file([self.relic_type, self.name, "html", name], html_path)
        self._add_metadata(metadata)

    # TODO needs test coverage
    def add_html_string(self, name: str, html_str: str):
        self.assert_valid_id(name)

        metadata = Metadata(name=name, data_type="html", relic=self.data)

        self.storage.put_text([self.relic_type, self.name, "html", name], html_str)

        self._add_metadata(metadata)

    def list_html(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "html"])

    def get_html(self, name: str) -> str:
        self.assert_valid_id(name)

        with self.storage.get_binary_obj(
            [self.relic_type, self.name, "html", name]
        ) as f:
            return f.read().decode("utf-8")

    def add_text(self, name: str, text: str) -> None:

        self.assert_valid_id(name)

        size = getsizeof(text)
        metadata = Metadata(
            name=name,
            data_type="text",
            relic=self.data,
            size=size,
            shape=len(text),
        )

        self.storage.put_text([self.relic_type, self.name, "text", name], text)
        self._add_metadata(metadata)

    def list_text(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "text"])

    def get_text(self, name: str) -> str:
        self.assert_valid_id(name)

        return self.storage.get_text([self.relic_type, self.name, "text", name])

    def _add_metadata(self, metadata: Metadata) -> None:
        self.assert_valid_id(metadata.name)

        self.storage.put_metadata(
            [self.relic_type, self.name, "metadata", metadata.data_type, metadata.name],
            metadata.get_dict(),
        )

    def describe(self) -> Dict:
        return self.storage.get_metadata(
            [self.relic_type, self.name, "metadata"], self.name
        )

    def add_tag(self, tags: Dict) -> None:
        tags_path = [self.relic_type, self.name, "tags"]

        curr_tags = self.storage.get_tags(tags_path)
        curr_tags.update(tags)

        self.storage.put_tags(tags_path, curr_tags)

    def list_tags(self) -> Dict:
        return self.storage.get_tags([self.relic_type, self.name, "tags"])

    def add_image(self, name: str, image_bytes: BytesIO):
        self.assert_valid_id(name)

        size = image_bytes.getbuffer().nbytes
        metadata = Metadata(
            name=name,
            data_type="images",
            relic=self.data,
            size=size,
        )

        self.storage.put_binary_obj(
            [self.relic_type, self.name, "images", name], image_bytes
        )
        self._add_metadata(metadata)

    def add_image_from_path(self, name: str, image_path: str):
        self.assert_valid_id(name)
        # TODO: Make use of stream like capabilities instead of full read()s
        with open(image_path, "rb") as input_file:
            buffer = BytesIO(input_file.read())
            fileSize = buffer.getbuffer().nbytes
            metadata = Metadata(
                name=name,
                data_type="images",
                relic=self._relic_data(),
                size=fileSize,
            )
            self.storage.put_binary_obj(
                [self.relic_type, self.name, "images", name], buffer
            )
            self._add_metadata(metadata)

    def get_image(self, name: str) -> BytesIO:
        self.assert_valid_id(name)
        return self.storage.get_binary_obj([self.relic_type, self.name, "images", name])

    def save_image_to_path(self, name: str, path: str) -> None:
        self.assert_valid_id(name)
        buffer = self.storage.get_binary_obj(
            [self.relic_type, self.name, "images", name]
        )
        content = buffer.read()
        with open(path, "wb") as new_file:
            new_file.write(content)

    def list_images(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "images"])

    def add_json(self, name: str, json_data: Dict) -> None:
        self.assert_valid_id(name)

        json_text = json.dumps(json_data)
        size = getsizeof(json_data)
        metadata = Metadata(
            name=name,
            data_type="json",
            relic=self._relic_data(),
            size=size,
        )

        self.storage.put_text([self.relic_type, self.name, "json", name], json_text)
        self._add_metadata(metadata)

    def list_json(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "json"])

    def get_json(self, name: str) -> Dict:
        self.assert_valid_id(name)

        json_text = self.storage.get_text([self.relic_type, self.name, "json", name])
        my_json = json.loads(json_text)
        return my_json

    def add_pandasdf(self, name: str, pandas_data: pd.DataFrame) -> None:
        """
        Note that json is used to serialize which
        comes with other caveats that can be found here:
        https://pandas.pydata.org/pandas-docs/version/0.23/generated/pandas.DataFrame.to_json.html
        """
        self.assert_valid_id(name)

        json_pandasdf = pandas_data.to_json()
        pandasdf_size = getsizeof(pandas_data)

        metadata = Metadata(
            name=name,
            data_type="pandasdf",
            relic=self._relic_data(),
            size=pandasdf_size,
        )

        self.storage.put_text(
            [self.relic_type, self.name, "pandasdf", name], json_pandasdf
        )
        self._add_metadata(metadata)

    def list_pandasdf(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "pandasdf"])

    def get_pandasdf(self, name: str) -> pd.DataFrame:
        self.assert_valid_id(name)

        pandas_json = self.storage.get_text(
            [self.relic_type, self.name, "pandasdf", name]
        )
        pandas_dataframe = pd.read_json(pandas_json)
        return pandas_dataframe

    def add_files_from_path(self, name: str, path: str) -> None:
        self.assert_valid_id(name)
        # TODO: Make use of stream like capabilities instead of full read()s
        with open(path, "rb") as input_file:
            buffer = BytesIO(input_file.read())
            fileSize = buffer.getbuffer().nbytes
            metadata = Metadata(
                name=name,
                data_type="files",
                relic=self._relic_data(),
                size=fileSize,
            )
            self.storage.put_binary_obj(
                [self.relic_type, self.name, "files", name], buffer
            )
            self._add_metadata(metadata)

    def save_files_to_path(self, name: str, path: str) -> None:
        buffer = self.storage.get_binary_obj(
            [self.relic_type, self.name, "files", name]
        )
        content = buffer.read()
        with open(path, "wb") as new_file:
            new_file.write(content)

    # TODO Add file like object

    def list_files(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "files"])

    def get_file(self, name: str) -> BytesIO:
        self.assert_valid_id(name)

        return self.storage.get_binary_obj([self.relic_type, self.name, "files", name])

    def add_notebook_from_path(self, name: str, path: str) -> None:
        self.assert_valid_id(name)
        # TODO: Make use of stream like capabilities instead of full read()s
        with open(path, "rb") as input_file:
            fileString = input_file.read()
            buffer = BytesIO(fileString)
            fileSize = buffer.getbuffer().nbytes
            metadata = Metadata(
                name=name,
                data_type="notebooks",
                relic=self._relic_data(),
                size=fileSize,
            )
            self.storage.put_binary_obj(
                [self.relic_type, self.name, "notebooks", name], buffer
            )
            self._add_metadata(metadata)

            exporter = nbconvert.HTMLExporter()
            exporter.template_name = "classic"
            note = nbformat.reads(fileString, as_version=4)
            (body, resources) = exporter.from_notebook_node(note)

            self.storage.put_text(
                [self.relic_type, self.name, "notebooks-html", name], body
            )

    def list_notebooks(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "notebooks"])

    def get_notebook(self, name: str) -> BytesIO:
        self.assert_valid_id(name)

        return self.storage.get_binary_obj(
            [self.relic_type, self.name, "notebooks", name]
        )

    def save_notebook_to_path(self, name: str, path: str) -> None:
        buffer = self.storage.get_binary_obj(
            [self.relic_type, self.name, "notebooks", name]
        )
        content = buffer.read()
        with open(path, "wb") as new_file:
            new_file.write(content)

    def get_notebook_html(self, name: str) -> str:
        self.assert_valid_id(name)

        return self.storage.get_text(
            [self.relic_type, self.name, "notebooks-html", name]
        )


class Reliquery:
    """
    Class used to query over available and accessible storage locations and Relics

    Attributes
    ----------
    storages : List[Storage]
        list of availably accessible storages

    metadata_db : MetadataDB
        in memory sqlite db used for querying Relics

    """

    def __init__(self, storages: List[Storage] = []) -> None:
        if len(storages) > 0:
            self.storages = storages
        else:
            self.storages = get_all_available_storages()

        self.storage_map = {s.name: s for s in self.storages}
        self.metadata_db = MetadataDB()
        self._sync_relics()

    def query(self, statement: str) -> List:
        return self.metadata_db.query(statement)

    def get_relics_by_tag(self, key: str, value: str) -> List[Relic]:
        """
        Query relics for user defined tags added to Relics.

        Parameters
        ----------

        key : string
        value : string

        Returns
        -------
        List of Relic objects
            Reclics associated with the key, values given
        """
        return [
            Relic(name=data[0], relic_type=data[1], storage=self.storage_map[data[2]])
            for data in self.metadata_db.get_relics_by_tag(key, value)
        ]

    def _sync_relics(self) -> None:
        """
        Syncs Relics from all available storages with the in-memory database.
        This is done on the initial creation of a Reliquery object and prior
        to any queries over Relics.
        """
        for stor in self.storages:
            relic_datas = [
                self.metadata_db.sync_relic_data(RelicData.parse_dict(data))
                for data in stor.get_all_relic_data()
            ]

            if relic_datas:
                for relic_data in relic_datas:
                    self.metadata_db.sync_tags(
                        RelicTag.parse_dict(
                            stor.get_tags(
                                [relic_data.relic_type, relic_data.relic_name, "tags"]
                            ),
                            relic_data,
                        )
                    )

    def get_relic_types_by_storage(self, storage: str) -> List[str]:
        return self.metadata_db.get_relic_types_by_storage(storage)

    def get_relic_names_by_storage_and_type(
        self, storage: str, relic_type: str
    ) -> List[Dict]:
        return self.metadata_db.get_relic_names_by_storage_and_type(storage, relic_type)

    def get_relic_names(self) -> List[Dict]:
        return self.metadata_db.get_all_relic_names()

    def get_unique_relic_types_and_storages(self) -> List[Dict]:
        return self.metadata_db.get_unique_relic_types_and_storages()

    def get_relic_names_by_storage(self, storage: str) -> List[Dict]:
        return self.metadata_db.get_relic_names_by_storage(storage)

    def sync_reliquery(self) -> None:
        self._sync_relics()



import logging
import os
from io import BytesIO, BufferedIOBase
from typing import Any, List, Dict
from shutil import copyfile
import json

import boto3
from botocore import UNSIGNED
from botocore.client import Config


from . import settings

import dropbox
from dropbox.exceptions import ApiError


StoragePath = List[str]

DATA_TYPES = [
    "arrays",
    "html",
    "text",
    "images",
    "json",
    "pandasdf",
    "files",
    "notebooks",
]


class StorageItemDoesNotExist(Exception):
    pass


class Storage:
    def put_file(self, path: StoragePath, file_path: str) -> None:
        raise NotImplementedError

    def put_binary_obj(self, path: StoragePath, buffer: BytesIO):
        raise NotImplementedError

    def get_binary_obj(self, path: StoragePath) -> BytesIO:
        raise NotImplementedError

    def put_text(self, path: StoragePath, text: str) -> None:
        raise NotImplementedError

    def get_text(self, path: StoragePath) -> str:
        raise NotImplementedError

    def list_keys(self, path: StoragePath) -> List[str]:
        raise NotImplementedError

    def put_metadata(self, path: StoragePath, metadata: Dict):
        raise NotImplementedError

    def get_metadata(self, path: StoragePath, root_key: str) -> Dict:
        raise NotImplementedError

    def put_tags(self, path: StoragePath, tags: Dict) -> None:
        raise NotImplementedError

    def get_tags(self, path: StoragePath) -> Dict:
        raise NotImplementedError

    def get_all_relic_tags(self) -> List[Dict]:
        raise NotImplementedError

    def get_all_relic_data(self) -> List[Dict]:
        raise NotImplementedError


class FileStorage(Storage):
    def __init__(self, root: str, name: str):
        self.root = os.path.expanduser(root)
        self.name = name

    def _ensure_path(self, path: StoragePath):
        if not os.path.exists(self._join_path(path[:-1])):
            os.makedirs(self._join_path(path[:-1]), exist_ok=True)

    def _join_path(self, path: StoragePath) -> str:
        return os.path.join(*[self.root] + path)

    def put_file(self, path: StoragePath, file_path: str) -> None:
        self._ensure_path(path)
        copyfile(file_path, self._join_path(path))

    def put_binary_obj(self, path: StoragePath, buffer: BytesIO) -> None:
        self._ensure_path(path)
        with open(self._join_path(path), "wb") as f:
            f.write(buffer.getbuffer())

    def get_binary_obj(self, path: StoragePath) -> BytesIO:
        return open(self._join_path(path), "rb")

    def put_text(self, path: StoragePath, text: str) -> None:
        self._ensure_path(path)

        with open(self._join_path(path), "w") as f:
            return f.write(text)

    def get_text(self, path: StoragePath) -> str:
        try:
            with open(self._join_path(path), "r") as f:
                return f.read()
        except IOError:
            raise StorageItemDoesNotExist

    def list_keys(self, path: StoragePath) -> List[str]:
        joined_path = self._join_path(path)
        if not os.path.exists(joined_path):
            return []

        if not os.path.isdir(joined_path):
            return []

        return os.listdir(joined_path)

    def put_metadata(self, path: StoragePath, metadata: Dict):
        self._ensure_path(path)

        with open(self._join_path(path), "w") as f:
            return f.write(json.dumps(metadata))

    def get_metadata(self, path: StoragePath, root_key: str) -> Dict:
        data = {root_key: {k: [] for k in DATA_TYPES}}

        def dict_from_path(path: StoragePath, dirname: str):
            dirpath = path.copy()
            dirpath.append(dirname)
            if os.path.exists(self._join_path(dirpath)):
                entries = os.listdir(self._join_path(dirpath))
                for i in entries:
                    entry_path = dirpath.copy()
                    entry_path.append(i)
                    with open(self._join_path(entry_path), "r") as f:
                        data[root_key][dirname].append(json.loads(f.read()))

        for d in data[root_key]:
            dict_from_path(path, d)

        return data

    def list_key_paths(self, path: StoragePath) -> List[str]:
        paths = []
        joined_path = self._join_path(path)
        subs = os.listdir(joined_path)

        if subs:
            for sub in subs:
                copy = path.copy()
                copy.append(sub)
                sub_path = self._join_path(copy)
                if os.path.isdir(sub_path):
                    paths.extend(self.list_key_paths(copy))
                else:
                    paths.append(sub_path)

        return paths

    def put_tags(self, path: StoragePath, tags: Dict) -> None:
        self._ensure_path(path)

        with open(self._join_path(path), "w") as f:
            return f.write(json.dumps(tags))

    def get_tags(self, path: StoragePath) -> Dict:
        try:
            with open(self._join_path(path), "r") as f:
                return json.loads(f.read())
        except FileNotFoundError:
            return {}

    def get_all_relic_tags(self) -> List[Dict]:
        tag_keys = [
            key for key in self.list_key_paths([""]) if "tags" in key.split("/")
        ]

        for key in tag_keys:
            return self.get_tags(key.split("/")[-3:])

    def get_all_relic_data(self) -> List[Dict]:
        relic_types = os.listdir(self._join_path([""]))
        relic_data = []
        for relic_type in relic_types:
            if os.path.isdir(self._join_path([relic_type])):
                names = os.listdir(self._join_path([relic_type]))

                if names:
                    for name in names:
                        relic_data.append(
                            {
                                "relic_name": name,
                                "relic_type": relic_type,
                                "storage_name": self.name,
                            }
                        )
        return relic_data


S3Client = Any


def _get_s3_client(signed) -> S3Client:
    if signed:
        return boto3.client("s3")
    else:
        return boto3.client("s3", config=Config(signature_version=UNSIGNED))


class S3Storage(Storage):
    def __init__(
        self,
        s3_bucket: str,
        prefix: str,
        name: str,
        s3_signed: bool = True,
    ):
        self.s3_bucket = s3_bucket
        self.prefix = prefix
        self.name = name if len(name) > 0 else s3_bucket
        self.signed = s3_signed

        self.s3 = _get_s3_client(self.signed)

    def _join_path(self, path: StoragePath) -> str:
        return "/".join([self.prefix] + path)

    def put_file(self, path: StoragePath, file_path: str) -> None:
        self.s3.upload_file(file_path, self.s3_bucket, self._join_path(path))

    def put_binary_obj(self, path: StoragePath, buffer: BufferedIOBase) -> None:
        self.s3.upload_fileobj(buffer, self.s3_bucket, self._join_path(path))

    def get_binary_obj(self, path: StoragePath) -> BufferedIOBase:
        buffer = BytesIO()

        self.s3.download_fileobj(self.s3_bucket, self._join_path(path), buffer)

        buffer.seek(0)

        return buffer

    def put_text(self, path: StoragePath, text: str, encoding: str = "utf-8") -> None:
        self.s3.put_object(
            Key=self._join_path(path), Bucket=self.s3_bucket, Body=text.encode(encoding)
        )

    def get_text(self, path: StoragePath, encoding: str = "utf-8") -> str:
        try:
            obj = self.s3.get_object(Key=self._join_path(path), Bucket=self.s3_bucket)
        except self.s3.exceptions.NoSuchKey:
            raise StorageItemDoesNotExist

        return obj["Body"].read().decode(encoding)

    def list_keys(self, path: StoragePath) -> List[str]:
        prefix = self._join_path(path)

        def process_key(k):
            return k[len(prefix) + 1 :]

        is_truncated = True
        keys = []
        kwargs = {}
        while is_truncated:
            response = self.s3.list_objects_v2(
                Bucket=self.s3_bucket, Prefix=prefix, **kwargs
            )

            is_truncated = response["IsTruncated"]
            if is_truncated:
                kwargs = dict(ContinuationToken=response["NextContinuationToken"])

            try:
                keys.extend([process_key(c["Key"]) for c in response["Contents"]])
            except KeyError:
                logging.warning(f"No files found in directory {prefix}")
        # we want to remove the prefx
        return keys

    def put_metadata(self, path: Storage, metadata: Dict):
        self.s3.put_object(
            Key=self._join_path(path), Bucket=self.s3_bucket, Body=json.dumps(metadata)
        )

    def get_metadata(self, path: StoragePath, root_key: str) -> Dict:
        data = {root_key: {k: [] for k in DATA_TYPES}}

        def dict_from_path(path: StoragePath, dirname: str) -> None:
            dirpath = path.copy()
            dirpath.append(dirname)
            dir_keys = self.list_keys(dirpath.copy())

            for key in dir_keys:
                key_path = dirpath.copy()
                key_path.append(key)

                try:
                    obj = self.s3.get_object(
                        Key=self._join_path(key_path),
                        Bucket=self.s3_bucket,
                    )
                except self.s3.exceptions.NoSuchKey:
                    raise StorageItemDoesNotExist

                data[root_key][dirname].append(
                    json.loads(obj["Body"].read().decode("utf-8"))
                )

        for d in data[root_key]:
            dict_from_path(path, d)

        return data

    def list_key_paths(self, path: StoragePath) -> List[str]:
        prefix = self._join_path(path)

        def process_key(k):
            return k[len(prefix) :]

        is_truncated = True
        keys = []
        kwargs = {}
        while is_truncated:
            response = self.s3.list_objects_v2(
                Bucket=self.s3_bucket, Prefix=prefix, **kwargs
            )

            is_truncated = response["IsTruncated"]
            if is_truncated:
                kwargs = dict(ContinuationToken=response["NextContinuationToken"])

            try:
                keys.extend([process_key(c["Key"]) for c in response["Contents"]])
            except KeyError:
                logging.warning(f"No files found in directory {prefix}")
        # we want to remove the prefx
        return keys

    def put_tags(self, path: StoragePath, tags: Dict) -> None:
        self.put_text(path, json.dumps(tags))

    def get_tags(self, path: StoragePath) -> Dict:
        try:
            return json.loads(self.get_text(path))
        except StorageItemDoesNotExist:
            return {}

    def get_all_relic_tags(self) -> List[Dict]:
        tag_keys = [
            key for key in self.list_key_paths([""]) if "tags" in key.split("/")
        ]
        tags = []
        for key in tag_keys:
            tags.extend(self.get_tags(key.split("/")))

        return tags

    def get_all_relic_data(self) -> List[Dict]:
        relic_types = {path.split("/")[0] for path in self.list_key_paths([""])}
        relic_data = []

        for relic_type in relic_types:
            names = {path.split("/")[1] for path in self.list_key_paths([relic_type])}
            if names:
                for name in names:
                    relic_data.append(
                        {
                            "relic_name": name,
                            "relic_type": relic_type,
                            "storage_name": self.name,
                        }
                    )
            return relic_data


class DropboxStorage(Storage):
    def __init__(
        self,
        access_token: str,
        prefix: str,
        name: str,
    ):
        self.dbx = dropbox.Dropbox(access_token)
        self.prefix = "/" + prefix
        self.name = name

    def _join_path(self, path: StoragePath) -> str:
        return "/".join([self.prefix] + path)

    def put_file(self, path: StoragePath, file_path: str) -> None:
        path = path.split()
        with open(file_path, "rb") as f:
            self.dbx.files_upload(f.read(), "/" + self._join_path(path), mute=True)

    def put_binary_obj(self, path: StoragePath, buffer: BytesIO) -> None:
        self.dbx.files_upload(buffer.read(), self._join_path(path), mute=True)

    def get_binary_obj(self, path: StoragePath) -> BytesIO:
        return BytesIO(
            self.dbx.files_download(self._join_path(path), rev=None)[-1].content
        )

    def put_text(self, path: StoragePath, text: str, encoding: str = "utf-8") -> None:
        string_bytes = bytes(text, encoding)
        self.dbx.files_upload(string_bytes, self._join_path(path))

    def get_text(self, path: StoragePath) -> str:
        try:
            return self.dbx.files_download(self._join_path(path))[-1].text
        except ApiError:
            raise StorageItemDoesNotExist

    def list_keys(self, path: StoragePath) -> List[str]:
        try:
            return [
                i.name
                for i in self.dbx.files_list_folder(self._join_path(path)).entries
            ]
        except ApiError:
            return []

    def put_metadata(self, path: StoragePath, metadata: Dict):
        metadata_bytes = json.dumps(metadata).encode("utf-8")
        self.dbx.files_upload(
            metadata_bytes,
            self._join_path(path),
            mode=dropbox.files.WriteMode.overwrite,
        )

    def get_metadata(self, path: StoragePath, root_key: str) -> Dict:
        data = {root_key: {k: [] for k in DATA_TYPES}}

        def dict_from_path(path: StoragePath, dirname: str):
            dirpath = path.copy()
            dirpath.append(dirname)
            entries = self.list_keys(dirpath)

            for i in entries:
                entry_path = dirpath.copy()
                entry_path.append(i)
                data[root_key][dirname].append(
                    self.dbx.files_download(self._join_path(entry_path))[-1].json()
                )

        for d in data[root_key]:
            dict_from_path(path, d)

        return data

    def list_key_paths(self, path: StoragePath) -> List[str]:
        paths = []
        subs = self.list_keys(path)
        if subs:
            for sub in subs:
                copy = path.copy()
                copy.append(sub)
                sub_path = self._join_path(copy)
                # isinstace was found as the best way to check if a directory
                # is a file or folder
                if isinstance(
                    self.dbx.files_get_metadata(sub_path), dropbox.files.FolderMetadata
                ):
                    paths.extend(self.list_key_paths(copy))
                elif isinstance(
                    self.dbx.files_get_metadata(sub_path), dropbox.files.FileMetadata
                ):
                    paths.append(sub_path)
        return paths

    def put_tags(self, path: StoragePath, tags: Dict) -> None:
        self.put_text(path, json.dumps(tags))

    def get_tags(self, path: StoragePath) -> Dict:
        try:
            return json.loads(self.get_text(path))
        except StorageItemDoesNotExist:
            return {}

    def get_all_relic_tags(self) -> List[Dict]:
        tag_keys = [key for key in self.list_key_paths([]) if "tags" in key.split("/")]

        for key in tag_keys:
            return self.get_tags(key.split("/")[-3:])

    def get_all_relic_data(self) -> List[Dict]:
        relic_types = {path.split("/")[2] for path in self.list_key_paths([])}
        relic_data = []

        for relic_type in relic_types:
            names = {path.split("/")[3] for path in self.list_key_paths([relic_type])}
            if names:
                for name in names:
                    relic_data.append(
                        {
                            "relic_name": name,
                            "relic_type": relic_type,
                            "storage_name": self.name,
                        }
                    )
            return relic_data


def get_storage_by_name(name: str, root: str = os.path.expanduser("~")) -> Storage:
    reliquery_dir = os.path.join(root, "reliquery")
    config = settings.get_config(reliquery_dir)

    return get_storage(name, reliquery_dir, config[name])


def get_all_available_storages(root: str = os.path.expanduser("~")) -> List[Storage]:
    reliquery_dir = os.path.join(root, "reliquery")
    config = settings.get_config(reliquery_dir)

    storages = []

    for key in config.keys():
        storages.append(get_storage(key, reliquery_dir, config[key]))

    return storages


def get_storage(name: str, root: str, config: Dict) -> Storage:
    if config["storage"]["type"] == "S3":
        return S3Storage(
            **config["storage"]["args"],
            name=name,
        )

    elif config["storage"]["type"] == "File":
        if "root" in config["storage"]["args"]:
            return FileStorage(**config["storage"]["args"], name=name)
        else:
            return FileStorage(**config["storage"]["args"], root=root, name=name)

    elif config["storage"]["type"] == "Dropbox":
        return DropboxStorage(**config["storage"]["args"], name=name)

    else:
        raise ValueError(f"No storage found by name : {name}")



from setuptools import setup

install_requires = [
    "numpy >= 1.16",
    "boto3 >= 1.17",
    "black>=21.4b",
    "pytest",
    "flake8",
    "sphinx",
    "pandas",
    "nbconvert",
    "nbformat",
    "dropbox",
]

setup(
    name="reliquery",
    version="0.3.2",
    description="Science's Artifact Antiformat",
    url="https://github.com/The-Dev-Effect/reliquery",
    author="The Dev Effect",
    author_email="company@thedeveffect.com",
    license="MIT OR Apache-2.0",
    packages=["reliquery"],
    tests_require=["pytest"],
    install_requires=install_requires,
)
